#ifndef _POWER_BB_H_
#define _POWER_BB_H_

#include "common/h/Vector.h"
#include "util/h/kdrvtypes.h"
#include <inttypes.h> // uint32_t
#include "power_instr.h"
#include "fnCode.h"
#include "powerTraits.h"
#include "power_bitwise_dataflow_fn.h"
// fwd decls to avoid recursive #includes:
class basicblock_t;
class power_reg_set;
class monotone_bitwise_dataflow_fn;

class power_bb {
public:
   typedef fnCode fnCode_t;
   typedef uint16_t bbid_t; //same as in basicblock.h

   static void
      factorInControlDependencyInRegsToSlice(regset_t  &  /* regsToSlice */ ,
                                             const fnCode_t::codeChunk  & /* theCodeChunk */,
                                             kptr_t  /* bbStartAddr  */,
                                             kptr_t /* bbEndAddrPlus1*/) {
      // This routine adds the PC register to regsToSlice if the bb ends with a
      // *conditional* branch.
      //Applicable to Power??
      assert(false && "not implemented until a need for such functionality appears");
   }

      static pdvector<bool> backwardsSlice(kptr_t /* succ_bb_startAddr*/, // used for ds on sparc
				      regset_t &regsToSlice,
				      bool includeControlDependencies,
				      const fnCode_t &theFnCode,
				      unsigned numInsnBytesToSkipAtEnd,
				      unsigned bbNumInsnBytes,
				      kptr_t bbStartAddr,
				      kptr_t bbEndAddrPlus1) {
     // numInsnBytesToSkipAtEnd: the number of instruction bytes out of this basic
     // block, counting from the bottom of the basic block, that should be ignored
     // when doing the slice.  Of course 0 means skip nothing and consider the
     // whole basic block when slicing.
     assert(numInsnBytesToSkipAtEnd <= bbNumInsnBytes);

   // In the future, we may allow slicing by memory locations, not just regs.

     fnCode_t::const_iterator chunk_iter = theFnCode.findChunk(bbStartAddr, false);
     assert(chunk_iter != theFnCode.end());
     const fnCode_t::codeChunk &theCodeChunk = *chunk_iter;

     const unsigned bbNumInsns = 
        theCodeChunk.calcNumInsnsWithinBasicBlock(bbStartAddr,
                                                  bbEndAddrPlus1);

     // initialize the result: no instructions are in the slice
     pdvector<bool> result(bbNumInsns, false);
     
     
     //   ****** Not applicable to Power?? -Igor ****************
     // add the PC register to regsToSlice if bb ends in a *conditional* branch (unless
     // it falls outside of the range of insns to slice thanks to numInsnBytesToSkipAtEnd)
     if (includeControlDependencies && numInsnBytesToSkipAtEnd == 0)
        factorInControlDependencyInRegsToSlice(regsToSlice, theCodeChunk, 
                                               bbStartAddr, bbEndAddrPlus1);
    


     unsigned lcv = bbNumInsns - 1 - (numInsnBytesToSkipAtEnd / 4);
     
     unsigned byteoffset = bbNumInsnBytes - 4 - numInsnBytesToSkipAtEnd;
     
     while (true) {
        if (regsToSlice.isempty())
           return result;
        
        assert(byteoffset < bbNumInsnBytes);
        const power_instr *instr = 
           (const power_instr *)theCodeChunk.get1Insn(bbStartAddr + byteoffset);
        
        power_instr::registerUsage ru;
        instr->getRegisterUsage(&ru);
        
        *(power_reg_set *)ru.definitelyUsedBeforeWritten -=  
           power_reg_set::regsThatIgnoreWrites;
        *(power_reg_set *)ru.maybeUsedBeforeWritten -= 
           power_reg_set::regsThatIgnoreWrites;

        const power_reg_set regsToSlice_definitelyWritten = 
           *(power_reg_set *)ru.definitelyWritten &  (power_reg_set &)regsToSlice;
        const power_reg_set regsToSlice_maybeWritten = 
           *(power_reg_set *)ru.maybeWritten & (power_reg_set &)regsToSlice;

           
        power_reg_set &regsToSlice2 = (power_reg_set &)regsToSlice;
        //allow for easy use of power_reg_set operators
        
        if (!regsToSlice_definitelyWritten.isempty()) {
           // The insn contributes to the value of some reg(s) we're slicing; thus, the
           // insn is part of the slice.
           result[lcv] = true;

           // The insn kills the set regsToSlice_definitelyWritten 
           //(probably just a single entry in the set, by the way).  Since 
           //they're dead, no previous insn can affect them insofar as the
           //slice concerns.  Our register interest now shifts to registers 
           //_used_ by this insn.
           
           //Note that we want to add in all the registers that have been
           //used here, not just the ones we're interested in, since
           //those other registers might contain the data from the
           //registers we are initially interested in. -Igor
        
           regsToSlice2 -= regsToSlice_definitelyWritten; // dead
           
           regsToSlice2 += *(power_reg_set *)ru.definitelyUsedBeforeWritten; // g0 has been removed -- good
           regsToSlice2 += *(power_reg_set *)ru.maybeUsedBeforeWritten; // g0 has been removed -- good
        }
        if (!regsToSlice_maybeWritten.isempty()) {
           // This instr *might* contribute to the value of some regs we care about.
           // Thus, add it to the slice.
           result[lcv] = true;
           
           // The regs written to cannot be declared dead.
           ;
           
           // Live regs which contribute to the slice: care about 'em:
           regsToSlice2 += *(power_reg_set *)ru.definitelyUsedBeforeWritten; // g0 has been removed -- good
           regsToSlice2 += *(power_reg_set *)ru.maybeUsedBeforeWritten; // g0 has been removed -- good
        }
        
        if (lcv == 0 && byteoffset == 0)
           break;
        else if (lcv == 0)
           assert(false); // should be both or neither
        else if (byteoffset == 0)
           assert(false); // should be both or neither
        
        byteoffset -= instr->getNumBytes();
        --lcv;
     }
     
     return result;
   }

   // Sun's CC says templated methods can't be static.  I disagree.
   template <class calcEffectOfCall>
      static monotone_bitwise_dataflow_fn *
      doLiveRegAnalysis(kptr_t addrWhereToStart, // such as just past the end of the bb
                        kptr_t addrWhereToStop, // such as the startAddr of the bb
                     const fnCode_t &theFnCode,
                     const calcEffectOfCall &callOracle,
                     unsigned bbNumInsnBytes,
		     kptr_t bbStartAddr,
		     kptr_t bbEndAddrPlus1,
		     unsigned bbNumSuccs,
			bbid_t bbSuccOf0);
 
};

#endif /* POWER_BB_H_ */
