// kerninst-x86.h - functionality only needed by x86 linux driver

#ifndef i386_unknown_linux2_4
#error "kerninst-x86.h should only be include for platform i386_unknown_linux2_4"
#endif

#include <asm/msr.h> // rdtsc()
#include <asm/desc.h>
#include <linux/version.h> // LINUX_VERSION_CODE

#include "hwcounter_p4.h"

extern struct cpuinfo_x86 boot_cpu_data;

/*--------------- Hardware Counter State ---------------*/

perfctr_state_t reset_perfctr_state = {
   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,0}
};

/*--------------- Trap-Based Instrumentation ---------------*/
void get_idt_base_addr(void);
void get_int3_handler(u_int);
int set_int3_handler(u_long);
u_long match_addr(u_long);
int install_trap_mapping(u_long, u_long);
int update_trap_mapping(u_long, u_long);
int uninstall_trap_mapping(u_long);

#define DOINT3_TYPE void
#define DOINT3_ATTR asmlinkage
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9)
#  ifdef CONFIG_KPROBES
#    undef DOINT3_TYPE
#    define DOINT3_TYPE int
#    define DOINT3_TYPE_INT
#  endif 
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
#  undef DOINT3_ATTR
#  define DOINT3_ATTR fastcall
#  ifdef CONFIG_KPROBES
#    if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
#      undef DOINT3_TYPE
#      define DOINT3_TYPE int
#      define DOINT3_TYPE_INT
#    endif
#  endif
#endif

DOINT3_ATTR DOINT3_TYPE kerninst_int3_handler(struct pt_regs *, long);
DOINT3_ATTR DOINT3_TYPE (*do_int3)(struct pt_regs *, long);

/* hash table of trap addr to patch addr mappings */
static struct hash_table ht_traps;

static int num_traps_installed = 0;
static u_long do_int3_addr = 0;
static struct desc_struct *idt_base = NULL;
static u_long int3_handler_addr = 0;
static unsigned char int3_handler_installed = 0;

void benchmark_instrumentation(void)
{
   unsigned long long start = 0, stop = 0, total = 0;
   unsigned exec_count = 1000;
   unsigned i = 0;
   for(; i < exec_count; i++) {
      rdtscll(start);
#if 1 /* profile traps */
      __asm__ __volatile__ ("addb $0, %%al\n" : 
                            /*no inputs*/ : 
                            /*no outputs*/);
#else /* profile branches */
      __asm__ __volatile__ ("addl $0x10000000, %%eax\n" : 
                            /*no inputs*/ : 
                            /*no outputs*/);
#endif
      rdtscll(stop);
      total += stop - start;
   }
   printk(KERN_INFO "kerninst: benchmark_instrumentation - total cycles at inst point is %lld for %d executions\n", total, exec_count);
}

void get_idt_base_addr(void)
{
   unsigned space[2];
   __asm__ __volatile__ ("sidt %0\n": :"m" (*((short *)space+1)));
   idt_base = (struct desc_struct *)(space[1]);
}

static unsigned char kprobes_int3 = 0;

void get_int3_handler(unsigned first)
{
   u_long part1, part2 = 0;
   struct desc_struct *temp;
   unsigned char pushed_val = 0;

   if( first ) {
      if( idt_base == NULL ) get_idt_base_addr();

      temp = idt_base + 3;
      part1 = temp->a;
      part2 = temp->b;
      int3_handler_addr = (part2 & 0xFFFF0000) + (part1 & 0x0000FFFF);

      pushed_val = *(((unsigned char*)int3_handler_addr)+1);
      if(pushed_val == 0xFF) {
	 kprobes_int3 = 1;
         /* kprobes patch has been applied, so address of do_int3 can be
            calculated from the call to it */
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9)
         /* call is at offset 0x1a from int3_handler_addr */
         do_int3_addr = *(u_long *)(((char*)int3_handler_addr)+0x1b) + 
                        int3_handler_addr + 0x1f;
#else
         /* call is at offset 0x19 from int3_handler_addr */
         do_int3_addr = *(u_long *)(((char*)int3_handler_addr)+0x1a) + 
                        int3_handler_addr + 0x1e;
#endif
      }
      else {
	 /* address of do_int3 is pushed in 2nd instruction, and located
            at offset 3 from int3_handler_addr */
         do_int3_addr = *(u_long *)(((char*)int3_handler_addr)+3);
      }
      do_int3 = (DOINT3_TYPE (*)(struct pt_regs *, long))do_int3_addr;
   }
}

int set_int3_handler(u_long addr)
{
   /* don't set if address given is null */
   if( (void *)addr == NULL ) return 1;
 
   if(kprobes_int3) {
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9)
      addr = addr - (int3_handler_addr + 0x1f);
      *(u_long *)(((char*)int3_handler_addr)+0x1b) = addr;
#else
      addr = addr - (int3_handler_addr + 0x1e);
      *(u_long *)(((char*)int3_handler_addr)+0x1a) = addr;
#endif
   }
   else
      *(u_long *)(((char*)int3_handler_addr)+3) = addr;
 
   if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
      printk(KERN_INFO "kerninst: set int3 handler to call %08lx\n", addr);
   return 0;
}

int install_trap_mapping(u_long trap_addr, u_long patch_addr)
{
   int ret;
 
   /* install new int3 handler if not already installed */
   if(! int3_handler_installed) {
      set_int3_handler((u_long)kerninst_int3_handler);
      int3_handler_installed = 1;
   }

   if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
      printk(KERN_INFO "kerninst: adding mapping for trap @ 0x%08lx, corresponding patch @ 0x%08lx\n", trap_addr, patch_addr);

   /* insert mapping into list */
   ret = hash_table_add(&ht_traps, (void*)patch_addr, trap_addr);
   if(ret == 0)
      num_traps_installed++;

   return ret;
}

int update_trap_mapping(u_long trap_addr, u_long patch_addr)
{
   int ret;
   if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
      printk(KERN_INFO "kerninst: updating mapping for trap @ 0x%08lx, new patch @ 0x%08lx\n", trap_addr, patch_addr);
   ret = hash_table_update(&ht_traps, trap_addr, (void*)patch_addr);
   return ret;
}

int uninstall_trap_mapping(u_long trap_addr)
{
   void *patch_addr;

   if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
      printk(KERN_INFO "kerninst: removing mapping for trap @ 0x%08lx\n", trap_addr);

   /* remove mapping from list, if it exists */
   patch_addr = hash_table_remove(&ht_traps, trap_addr);
   if(patch_addr != NULL) {
      num_traps_installed--;

      /* uninstall int3 handler if there are no more traps */
      if(! num_traps_installed) {
         if( int3_handler_installed ) {
            set_int3_handler(do_int3_addr);
            int3_handler_installed = 0;
         }
      }
      return 0;
   }
   return -1;
}

u_long match_addr(u_long trap_addr)
{
   void *patch_addr = hash_table_lookup(&ht_traps, trap_addr);
   return (u_long)patch_addr;
}

DOINT3_ATTR DOINT3_TYPE kerninst_int3_handler(struct pt_regs *regs, long error_code)
{
   u_long patch_addr = 0;
   u_long trap_addr = (u_long)regs->eip - 1; // '-1' since int3 only one byte
   if( (!(regs->xcs & 3)) && 
       ((patch_addr = match_addr(trap_addr)) != 0) ) {
      /* code segment specifes kernel & lookup of trap_addr successful */
      if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
         printk(KERN_INFO "kerninst: handling trap @ 0x%08lx, corresponding patch @ 0x%08lx\n", trap_addr, patch_addr);

      regs->eip = patch_addr; /* return to patch, not trap_addr+1 */
#ifdef DOINT3_TYPE_INT
      return 0;
#endif
   }
   else { /* not our stuff, call original do_int3 handler */
      if(kerninst_debug_bits & KERNINST_DEBUG_TRAPS)
         printk(KERN_INFO "kerninst: passing on int3 @ 0x%08lx to original handler\n", trap_addr);
#ifdef DOINT3_TYPE_INT
      return
#endif
      do_int3(regs, error_code);
   }
}
