/* symbol_table.c */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/kallsyms.h>
#include <asm/uaccess.h>
#include "symbol_table.h"

unsigned kerninst_symtab_do(int fillin, unsigned char *buf)
{
   unsigned long nbytes = 0;

   const char *name = NULL;
   unsigned char nmods = 0, name_len = 0;
   unsigned long name_offset = 0;
  
   unsigned char *bufp = buf;
   struct module *start = THIS_MODULE;
   struct module *modp = start;  
  
#ifndef CONFIG_KALLSYMS
   struct kernel_symbol *ksymp = NULL, *gpl_ksymp = NULL; 
   unsigned int num_ksyms = 0, num_gpl_ksyms = 0;
#endif

   unsigned char *str_tab_start = NULL;
   unsigned long str_tab_len, str_tab_len_offset;

   Elf_Sym *kallsymp = NULL;
   unsigned long num_kallsyms;

   struct kerninst_symbol ksym;

   struct list_head *modules_head = NULL, *modules_iter = NULL;

   /* sanity check */
   if( fillin && !bufp ) return 0;

   /* kludge to find head of modules list */
   modules_iter = modp->list.next;
   while(modules_iter != &(modp->list)) {
      modp = list_entry(modules_iter, struct module, list);
      if((modp->state != MODULE_STATE_LIVE) &&
         (modp->state != MODULE_STATE_COMING) &&
         (modp->state != MODULE_STATE_GOING)) {
      	 modules_head = modules_iter;
	 break;
      }
      modules_iter = modules_iter->next;
   }
   if(modules_head == NULL) return nbytes;
   
   /* record number of modules, including kernel */
   modp = start;
   list_for_each_entry(modp, modules_head, list) {
      nmods++;
   }
   
   if(fillin) { 
       *(unsigned char *)(buf+nbytes) = nmods + 1; /* +1 for "kernel" module */
   }  
   nbytes += sizeof(unsigned char); /* to hold number of modules */

   /* record info on each module */
   modp = start;
   list_for_each_entry(modp, modules_head, list) {
      name = modp->name;
      /* get module name information */
      name_len = strlen(name) + 1;
      if(fillin) {
         *(unsigned char*)(buf + nbytes) = name_len;
         strncpy(buf + nbytes + sizeof(unsigned char), name, name_len);
      }
      nbytes += sizeof(unsigned char);
      nbytes += name_len;
#ifdef ppc64_unknown_linux2_6
     nbytes = roundup(nbytes, sizeof(kptr_t));
#endif
      if(fillin) 
         *(kptr_t *)(buf + nbytes) = (kptr_t)modp->module_core;
      nbytes += sizeof(kptr_t);
      if (fillin)
         *(uint32_t*)(buf + nbytes) = modp->core_text_size;
      nbytes += sizeof(kptr_t);
      if (fillin)
         *(kptr_t *)(buf + nbytes) = (kptr_t)((char *)modp->module_core + modp->core_text_size);
      nbytes += sizeof(kptr_t);
      if (fillin)
         *(uint32_t*)(buf + nbytes) = modp->core_size - modp->core_text_size;
      nbytes += sizeof(kptr_t);
#ifdef ppc64_unknown_linux2_6
      /* toc section info, leave empty for now */
      if (fillin)
         *(kptr_t *)(buf + nbytes) = 0;
      nbytes += sizeof(kptr_t);
      if (fillin)
         *(uint32_t*)(buf + nbytes) = 0;
      nbytes += sizeof(kptr_t);
#endif
      
      /* loop through all symbols to find size of string table */
      str_tab_len = 0;
      str_tab_len_offset = nbytes;
      nbytes += sizeof(uint32_t);

#ifdef CONFIG_KALLSYMS
      num_kallsyms = modp->num_symtab;
      kallsymp = modp->symtab;
      str_tab_start = modp->strtab;
      while(num_kallsyms--) {
         name_len = strlen(str_tab_start + kallsymp->st_name) + 1;
         str_tab_len += name_len;
         if(fillin) {
            strncpy(buf + nbytes, str_tab_start + kallsymp->st_name, name_len);
         }
         nbytes += name_len;
         if(num_kallsyms)
            kallsymp++;
      }

      if(fillin) 
	*(uint32_t *)(buf + str_tab_len_offset) = str_tab_len;
#ifdef ppc64_unknown_linux2_6
        nbytes = roundup(nbytes, sizeof(kptr_t));
#endif
      /* a cookie marking the beginning of symbols */
      if (fillin) {
         *(unsigned*)(buf + nbytes) = 0x12345678;
      }
      nbytes += sizeof(unsigned);

      /* now process symbols */
    
      num_kallsyms = modp->num_symtab;
      if(fillin) {
         *(uint32_t*)(buf + nbytes) = num_kallsyms;
      }
      nbytes += sizeof(uint32_t); /* to hold number of symbols */
     
      kallsymp = modp->symtab;
      name_offset = 0;
      while(num_kallsyms--) {
	  if (fillin ) {
	      ksym.sym_addr = kallsymp->st_value;
	      ksym.sym_name_offset = name_offset;
	      memcpy(buf + nbytes, &ksym, sizeof(struct kerninst_symbol));
	  }
          name_len = strlen(str_tab_start + kallsymp->st_name);
          name_offset += name_len + 1;   
          nbytes += sizeof(struct kerninst_symbol);
#ifdef ppc64_unknown_linux2_4
	      nbytes = roundup(nbytes, sizeof(kptr_t));
#endif
	      
	  if(num_kallsyms)
	      kallsymp++;
      }
      if(name_offset != str_tab_len)
	  printk(KERN_WARNING "mismatch on string table\n");
      

#else // !CONFIG_KALLSYMS
      num_ksyms = modp->num_syms;
      ksymp = modp->syms;
      while(num_ksyms--) {
         name_len = strlen(ksymp->name) + 1;
         str_tab_len += name_len;
         if(fillin) {
            strncpy(buf + nbytes, ksymp->name, name_len);
	 }
	 nbytes += name_len;
         
         if(num_ksyms)
            ksymp++;
      }

      num_gpl_ksyms = modp->num_gpl_syms;
      gpl_ksymp = modp->gpl_syms;
      while(num_gpl_ksyms--) {
         name_len = strlen(gpl_ksymp->name) + 1;
         str_tab_len += name_len;
         if(fillin) {
            strncpy(buf+nbytes, gpl_ksymp->name, name_len);
	 }
	 nbytes += name_len;
         
         if(num_gpl_ksyms)
            gpl_ksymp++;
      }
      if(fillin) 
	*(uint32_t *)(buf + str_tab_len_offset) = str_tab_len;

      /* a cookie marking the beginning of symbols */
      if (fillin) {
         *(unsigned*)(buf + nbytes) = 0x12345678;
      }
      nbytes += sizeof(unsigned);

      /* now process symbols */
      num_kallsyms = modp->num_syms + modp->num_gpl_syms;
      if(fillin) {
	  *(uint32_t *)(buf + nbytes) = num_kallsyms;
      }
      nbytes += sizeof(uint32_t); /* to hold number of symbols */
      
      
      ksymp = modp->syms;
      num_ksyms = modp->num_syms;
      name_offset = 0;
      while(num_ksyms--) {
         if(fillin) {
            ksym.sym_addr = ksymp->value;
            ksym.sym_name_offset = name_offset;
            memcpy(buf + nbytes, &ksym, sizeof(struct kerninst_symbol));
         }
         name_len = strlen(ksymp->name);
         name_offset += name_len + 1;
         nbytes += sizeof(struct kerninst_symbol);
#ifdef ppc64_unknown_linux2_4
	 nbytes = roundup(nbytes, sizeof(kptr_t));
#endif
         if(num_ksyms)
            ksymp++;
      }
      gpl_ksymp = modp->gpl_syms;
      num_gpl_ksyms = modp->num_gpl_syms;
      while(num_gpl_ksyms--) {
         if( fillin) {      
            ksym.sym_addr = gpl_ksymp->value;
            ksym.sym_name_offset = name_offset;
            memcpy(buf + nbytes, &ksym, sizeof(struct kerninst_symbol));
         }
         name_len = strlen(gpl_ksymp->name);
         name_offset += name_len + 1;
         nbytes += sizeof(struct kerninst_symbol);
#ifdef ppc64_unknown_linux2_4
	 nbytes = roundup(nbytes, sizeof(kptr_t));
#endif
         if(num_gpl_ksyms)
            gpl_ksymp++;
      }
      if(name_offset != str_tab_len)
	  printk(KERN_WARNING "mismatch on string table\n");
      
#endif // CONFIG_KALLSYMS
   }

   /* record info on "kernel" module */

   
   /* module name */
   if (fillin)
       *(unsigned char*)(buf + nbytes)  = 7;
   nbytes += sizeof(unsigned char);
   if (fillin)
       strncpy(buf + nbytes, "kernel", 7);
   nbytes += 7;
#ifdef ppc64_unknown_linux2_4
   nbytes = roundup(nbytes, sizeof(kptr_t));
#endif 

   /* text and data sections */
   if (fillin)
       *(kptr_t *)(buf + nbytes) = (kptr_t)0;
   nbytes += sizeof(kptr_t);
   if (fillin)
       *(uint32_t*)(buf + nbytes) = 0;
   nbytes += sizeof(kptr_t);
   if (fillin)
       *(kptr_t *)(buf + nbytes) = (kptr_t)0;
   nbytes += sizeof(kptr_t);
   if (fillin)
       *(uint32_t*)(buf + nbytes) = 0;
   nbytes += sizeof(kptr_t);
#ifdef ppc64_unknown_linux2_6
   /* toc seciton */
   if (fillin)
       *(kptr_t *)(buf + nbytes) = (kptr_t)0;
   nbytes += sizeof(kptr_t);
   if (fillin)
       *(uint32_t*)(buf + nbytes) = 0;
   nbytes += sizeof(kptr_t);
#endif

      /* empty string table */
   if (fillin)
       *(uint32_t*)(buf + nbytes) = 0;
   nbytes += sizeof(kptr_t);

   /* cookie & no symbols */
   if (fillin)
       *(unsigned*)(buf + nbytes) = 0x12345678;
   nbytes += sizeof(unsigned);
   if (fillin)
       *(uint32_t*)(buf + nbytes) = 0;
   nbytes += sizeof(uint32_t);

   return nbytes;
}

static unsigned cached_symtab_size = 0;

unsigned kerninst_symtab_size(void)
{
   cached_symtab_size =  kerninst_symtab_do(0, (unsigned char *)NULL);
   return cached_symtab_size;
}

unsigned kerninst_symtab_get(kptr_t user_buf, int buf_size)
{
   void *buffer;
   unsigned result;
   unsigned using_vmalloc = 0;
   u_long pages, order = 0, base = 1;
   if( buf_size < cached_symtab_size ) {
      printk(KERN_WARNING "kerninst_symtab_get: buffer of size %d bytes too small to hold symbol table info\n", buf_size);
      return 0;
   }
   pages = (buf_size / 4096) + ( buf_size % 4096 ? 1 : 0 );
   while( base < pages ) {
      base <<= 1;
      order++;
   }
   buffer = (void *)__get_free_pages(GFP_KERNEL, order);
   if(!buffer) {
      using_vmalloc = 1;
      buffer = vmalloc(buf_size);
      if(!buffer) {
         printk(KERN_WARNING "kerninst_symtab_get: could not allocate in-kernel buffer of size %d for symbol table info\n", buf_size);
         return 0;
      }
   }
   copy_from_user(buffer, (void *)user_buf, (4 * sizeof(kptr_t)));
   result = kerninst_symtab_do(1, (unsigned char *)buffer);
   copy_to_user((void *)user_buf, buffer, result);
   if(using_vmalloc)
      vfree(buffer);
   else
      free_pages((u_long)buffer, order);
   return result;
}
